# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
GateImplementations for implementing quantum gates.

The :class:`.GateImplementation` subclasses in this subpackage construct :class:`.TimeBox` instances to
implement specific native gates, using the calibration data that the class has been initialized with.
Each GateImplementation instance encapsulates the calibration data for a specific implementation of a specific
native gate acting on a specific locus.

Several different implementations and calibration schemes can be supported for a given gate,
each represented by its own GateImplementation subclass.
Likewise, a single GateImplementation subclass can be sometimes used to implement several different gates
through different calibration data.
"""
from dataclasses import replace

import numpy as np

from iqm.pulse.gate_implementation import GateImplementation
from iqm.pulse.gates.barrier import Barrier
from iqm.pulse.gates.conditional import (
    CCPRX_Composite,
    CCPRX_Composite_DRAGCosineRiseFall,
    CCPRX_Composite_DRAGGaussian,
)
from iqm.pulse.gates.cz import (
    CZ_CRF,
    CZ_CRF_ACStarkCRF,
    CZ_GaussianSmoothedSquare,
    CZ_Slepian,
    CZ_Slepian_ACStarkCRF,
    CZ_Slepian_CRF,
    CZ_TruncatedGaussianSmoothedSquare,
    FluxPulseGate_CRF_CRF,
    FluxPulseGate_TGSS_CRF,
)
from iqm.pulse.gates.measure import Measure_Constant
from iqm.pulse.gates.move import MOVE_CRF_CRF, MOVE_TGSS_CRF
from iqm.pulse.gates.prx import (
    PRX_DRAGCosineRiseFall,
    PRX_DRAGCosineRiseFallSX,
    PRX_DRAGGaussian,
    PRX_DRAGGaussianSX,
    get_unitary_prx,
)
from iqm.pulse.gates.reset import Reset_Conditional, Reset_Wait
from iqm.pulse.gates.rz import RZ_ACStarkShift_CosineRiseFall, RZ_Virtual, get_unitary_rz
from iqm.pulse.gates.sx import SXGate
from iqm.pulse.gates.u import UGate, get_unitary_u
from iqm.pulse.quantum_ops import QuantumOp, QuantumOpTable

_exposed_implementations: dict[str, type[GateImplementation]] = {
    cls.__name__: cls  # type: ignore[misc]
    for cls in (
        Barrier,
        PRX_DRAGGaussian,
        PRX_DRAGCosineRiseFall,
        PRX_DRAGGaussianSX,
        PRX_DRAGCosineRiseFallSX,
        SXGate,
        UGate,
        RZ_Virtual,
        CZ_CRF_ACStarkCRF,
        CZ_Slepian_ACStarkCRF,
        CZ_GaussianSmoothedSquare,
        CZ_Slepian,
        CZ_Slepian_CRF,
        CZ_CRF,
        CZ_TruncatedGaussianSmoothedSquare,
        FluxPulseGate_TGSS_CRF,
        FluxPulseGate_CRF_CRF,
        Measure_Constant,
        MOVE_CRF_CRF,
        MOVE_TGSS_CRF,
        RZ_ACStarkShift_CosineRiseFall,
        CCPRX_Composite,
        CCPRX_Composite_DRAGCosineRiseFall,
        CCPRX_Composite_DRAGGaussian,
        Reset_Conditional,
    )
}
"""These GateImplementations can be referred to in the configuration YAML."""


def get_implementation_class(class_name: str) -> type[GateImplementation] | None:
    """Get gate implementation class by class name."""
    return _exposed_implementations.get(class_name, None)


def expose_implementation(implementation: type[GateImplementation], overwrite: bool = False) -> None:
    """Add the given gate implementation to the list of known implementations.

    Args:
        implementation: GateImplementation to add so that it can be found with :func:`.get_implementation_class`.
        overwrite: If True, does not raise an error if implementation already exists.
    """
    name = implementation.__name__
    if name in _exposed_implementations:
        if not overwrite and _exposed_implementations[name] is not implementation:
            raise ValueError(f"GateImplementation '{name}' has already been defined.")
    _exposed_implementations[name] = implementation


def register_implementation(
    operations: dict[str, QuantumOp],
    gate_name: str,
    impl_name: str,
    impl_class: type[GateImplementation],
    set_as_default: bool = False,
    overwrite: bool = False,
    quantum_op_specs: QuantumOp | dict | None = None,
) -> None:
    """Register a new gate implementation, and a new gate if needed.

    TODO: split the method for adding a new gate implementation and a new gate + implementation

    Args:
        operations: Known operations, mapping gate names to QuantumOps. A new QuantumOp is inserted here.
        gate_name: The gate name for which to register a new implementation.
        impl_name: The "human-readable" name with which the new gate implementation will be found e.g. in settings.
        impl_class: The python class of the new gate implementation to be addded.
        set_as_default: Whether to set the new implementation as the default implementation for the gate.
        overwrite: If True, allows replacing any existing implementation of the same name.
        quantum_op_specs: The quantum operation this gate represents. If a QuantumOp is given, it is used as is.
            If None is given and the same gate has been registered before, the previously registered properties are
            used.
            Otherwise, the given dict values are given to the constructor of :class:`~iqm.pulse.quantum_ops.QuantumOp`.
            For any missing constructor values, some defaults suitable for a 1-QB gate are used.
    """
    if isinstance(quantum_op_specs, QuantumOp):
        new = quantum_op_specs
    elif quantum_op_specs is None and gate_name in operations:
        new = operations[gate_name]
    else:
        new_kwargs = {
            "name": gate_name,
            "arity": 1,
            "params": tuple(),
            "implementations": {},
            "symmetric": impl_class.symmetric,
            "factorizable": False,
        }
        if quantum_op_specs:
            new_kwargs |= quantum_op_specs
            # cast iterables to tuple for consistency
            # TODO: fix the QuantumOp datclass comparisons and field typings properly
            new_kwargs["params"] = tuple(new_kwargs["params"])
        new = QuantumOp(**new_kwargs)

    # TODO: fix the QuantumOp datclass comparisons and field typings properly
    new_without_impls = replace(new, params=tuple(new.params), implementations={}, defaults_for_locus={})
    if gate_name in _default_operations:
        default_op = _default_operations[gate_name]
        default_without_impls = replace(default_op, implementations={}, defaults_for_locus={})
        if new_without_impls != default_without_impls:
            raise ValueError(
                f"{gate_name} is one of the default operation defined in iqm-pulse where it has different "
                f"properties. {default_op=}, {new=}"
            )

    if not overwrite and gate_name in operations:
        old = operations[gate_name]
        # TODO: fix the QuantumOp datclass comparisons and field typings properly
        old_without_impls = replace(old, params=tuple(old.params), implementations={}, defaults_for_locus={})
        if new_without_impls != old_without_impls:
            raise ValueError(f"{gate_name} has already been registered with different properties. {old=}, {new=}")

    new.implementations[impl_name] = impl_class
    if set_as_default and len(new.implementations) >= 1:
        new.set_default_implementation(impl_name)
    if not get_implementation_class(impl_class.__name__):
        expose_implementation(impl_class, overwrite)

    operations[gate_name] = new


_default_operations: QuantumOpTable = {
    op.name: op
    for op in [
        QuantumOp(
            "barrier",
            0,
            implementations={"": Barrier},
            symmetric=True,
        ),
        QuantumOp(
            "measure",
            0,
            ("key",),
            implementations={
                "constant": Measure_Constant,
                "constant_qnd": Measure_Constant,
            },
            factorizable=True,
        ),
        QuantumOp(
            "prx",
            1,
            ("angle", "phase"),
            implementations={
                "drag_gaussian": PRX_DRAGGaussian,
                "drag_crf": PRX_DRAGCosineRiseFall,
                "drag_crf_sx": PRX_DRAGCosineRiseFallSX,
                "drag_gaussian_sx": PRX_DRAGGaussianSX,
            },
            unitary=get_unitary_prx,
        ),
        QuantumOp(
            "u",
            1,
            ("theta", "phi", "lam"),
            implementations={"prx_u": UGate},
            unitary=get_unitary_u,
        ),
        QuantumOp("sx", 1, implementations={"prx_sx": SXGate}, unitary=lambda: get_unitary_prx(np.pi / 2, 0)),
        QuantumOp(
            "rz",
            1,
            ("angle",),
            implementations={"virtual": RZ_Virtual},
            unitary=get_unitary_rz,
        ),
        QuantumOp(
            "rz_physical",
            1,
            implementations={"ac_stark_crf": RZ_ACStarkShift_CosineRiseFall},
        ),
        QuantumOp(
            "cz",
            2,
            implementations={
                "tgss": CZ_TruncatedGaussianSmoothedSquare,
                "tgss_crf": FluxPulseGate_TGSS_CRF,
                "crf_crf": FluxPulseGate_CRF_CRF,
                "crf": CZ_CRF,
                "gaussian_smoothed_square": CZ_GaussianSmoothedSquare,
                "slepian": CZ_Slepian,
                "slepian_crf": CZ_Slepian_CRF,
                "crf_acstarkcrf": CZ_CRF_ACStarkCRF,
                "slepian_acstarkcrf": CZ_Slepian_ACStarkCRF,
            },
            symmetric=True,
            unitary=lambda: np.diag([1.0, 1.0, 1.0, -1.0]),
        ),
        QuantumOp(
            "move",
            2,
            implementations={
                "tgss_crf": MOVE_TGSS_CRF,
                "crf_crf": MOVE_CRF_CRF,
            },
        ),
        QuantumOp(
            "cc_prx",
            1,
            ("angle", "phase", "feedback_qubit", "feedback_key"),
            implementations={
                "prx_composite": CCPRX_Composite,
                "prx_composite_drag_crf": CCPRX_Composite_DRAGCosineRiseFall,
                "prx_composite_drag_gaussian": CCPRX_Composite_DRAGGaussian,
            },
        ),
        QuantumOp(
            "reset",
            0,
            implementations={"reset_conditional": Reset_Conditional},
            symmetric=True,
            factorizable=True,
        ),
        QuantumOp(
            "reset_wait",
            0,
            implementations={"reset_wait": Reset_Wait},
            symmetric=True,
            factorizable=True,
        ),
    ]
}
"""Native quantum operations are hardcoded here. They can be overridden using the YAML config file."""
# TODO for now, the default ops are hardcoded here, but they could always be read from a config file.



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for merging instruction schedules in time.

Under the idealized, noiseless, crosstalk-less computational model, the scheduling algorithms
should preserve the effect of the schedules on the computational subspace, i.e.
executing :class:`.Schedule` ``A`` immediately followed by ``B`` should be equivalent to
executing the merged schedule ``A+B``.

The merging is always done so that in ``A+B`` all the channels of ``B`` start
their execution simultaneously, and remain in sync. :class:`.Nothing` instructions can be added
as spacers between the channels of ``A`` and ``B`` as necessary to make this happen.

Typically the scheduling algorithms also try to minimize the total duration of the merged schedule.
"""
from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Optional

from iqm.pulse.playlist.channel import ChannelProperties
from iqm.pulse.playlist.instructions import Block, Instruction, Wait
from iqm.pulse.playlist.schedule import TOLERANCE, TOLERANCE_DECIMALS, Nothing, Schedule

# pylint: skip-file
# TODO: remove the above skips once COMP-1281 is done and the TETRIS mode is fixed.


NONSOLID = (Block, Nothing)
"""Instructions that can be converted to :class:`.Wait` after scheduling."""

BLOCKING = (Block,)  # TODO VirtualRZ, but we would need to keep track of the phase angles...
"""Nonsolid Instructions that nevertheless block."""


@dataclass
class SegmentPointer:
    """Tool for working with Segments.

    Serves as a time pointer inside the Segment.
    """

    source: list[Instruction]
    """contents of the Segment"""
    idx: int
    """index of the current instruction"""
    TOL: float
    """time durations equal or smaller than this tolerance are considered zero (in seconds)"""
    frac: float = 0.0  # assumed in ~[0, T)
    """time, in seconds, after the start of the current instruction"""

    def get(self) -> Instruction:
        """Current instruction."""
        return self.source[self.idx]

    @property
    def remainder(self) -> float:
        """Remaining duration of the current instruction (in seconds)."""
        return self.get().duration - self.frac

    def next(self) -> bool:
        """Move to the beginning of the next instruction."""
        self.idx += 1
        self.frac = 0.0
        return self.idx >= len(self.source)  # did we run out?

    def cut_tail(self) -> None:
        """Cut the source of the pointer at the current index.

        Truncate :attr:`source` at :attr:`idx`, make ``self`` point to
        the cut tail part only. :attr:`frac` is not changed.

        Note: modifies :attr:`source`.
        """
        tail = self.source[self.idx :]  # copy of the tail
        del self.source[self.idx :]  # remove tail from source, should be fast since it's the tail of a list?
        self.idx = 0
        self.source = tail

    def tail(self) -> Sequence[Instruction]:
        """Instructions from the current index onwards."""
        return self.source[self.idx :]

    def rewind(self, duration: float) -> None:
        """Move the pointer back by ``duration`` seconds."""
        while duration > self.TOL:
            delta = min(duration, self.frac)
            duration -= delta
            self.frac -= delta
            if self.frac < self.TOL:  # move generously, or we'll never get anywhere
                self.idx -= 1
                T = self.get().duration
                self.frac += T
        # finally, normalize the position
        if self.frac + self.TOL > self.get().duration:
            self.idx += 1
            self.frac = 0.0
        if self.idx < 0:
            raise IndexError("rewinded too far")

    def fastforward(self, duration: float) -> bool:
        """Move the pointer forward by ``duration`` seconds."""
        T = self.get().duration
        if duration <= self.TOL:
            if T < self.TOL:  # the current instruction is 0 duration, so we forward past that.
                self.idx += 1
            if self.idx >= len(self.source):  # ran out
                return True
        while duration > self.TOL:
            inst = self.get()
            if not isinstance(inst, NONSOLID):
                raise ValueError(f"trying to fastforward over solid things! {inst}")
            delta = min(duration, T - self.frac)
            duration -= delta
            self.frac += delta
            if self.frac + self.TOL > T:  # move generously
                self.idx += 1
                if self.idx >= len(self.source):
                    return True  # ran out
                self.frac -= T
                T = self.get().duration
        return False


def extend_schedule(
    A: Schedule,
    B: Schedule,
    channels: dict[str, ChannelProperties],
    *,
    TOL: float = TOLERANCE,
) -> None:
    """Extend a Schedule with another Schedule.

    Extends ``A`` with ``B``, modifying both. The extension can add new channels to ``A``.

    If ``B`` has a ragged left side, i.e. some of its channels begin with :class:`.Nothing` instructions,
    this algorithm will not always produce an optimally short schedule.

    Args:
        A: schedule to be extended
        B: schedule to extend ``A`` with
        channels: properties of the control channels
        TOL: time durations equal or smaller than this are considered zero (in seconds)
    """
    # TODO consequent VirtualRZ instructions could be merged
    # extend_schedule(A, A) causes issues if we do not make a copy
    B = B.copy()

    # add missing channels
    A.add_channels(tuple(B))

    start_times = {}  # earliest possible start times for each channel of B
    extra_block = {}  # blocking time after B_start_time for each channel
    for ch, segment_B in B.items():
        segment_A = A[ch]
        T = segment_A.duration
        if segment_A and isinstance(segment_A[-1], Block):
            # remove the block, record its duration
            block_A = segment_A.pop().duration
        else:
            block_A = 0

        if segment_B and isinstance(segment_B[0], Block):
            # remove the block, record its duration
            block_B = segment_B.pop(0).duration
            extra_block[ch] = block_B
            if not segment_B:
                # there is nothing else in segment_B
                block_B = float("inf")
        else:
            block_B = 0

        # Blocks can overlap (we assume there are no consequent Blocks)
        start_times[ch] = T - min(block_A, block_B)

    B_start_time = max(start_times.values(), default=0)

    # add the instructions from B to A, with an appropriate Block between them on each channel
    for ch, segment_B in B.items():
        segment_A = A[ch]
        duration = B_start_time - segment_A.duration + extra_block.get(ch, 0.0)
        if duration > TOL:
            # make sure the channel can handle the wait
            _ = channels[ch].duration_to_int_samples(duration, message=f"{ch}: Block duration")
            # TODO if a wait is shorter than the shortest the instruments can handle,
            # maybe make all the segments wait longer?
            segment_A.append(Block(duration))

        # add the instructions in B
        segment_A.extend(segment_B)


def extend_schedule_new(
    A: Schedule,
    B: Schedule,
    channels: dict[str, ChannelProperties],
    *,
    TOL: float = TOLERANCE,
) -> None:
    """Extend a Schedule with another Schedule.

    Extends ``A`` with ``B``, modifying ``A``. The extension can add new channels to ``A``.

    Can also handle cases where ``B`` has a ragged left side, i.e. some of its channels begin
    with :class:`.Nothing` instructions.

    Args:
        A: schedule to be extended
        B: schedule to extend ``A`` with
        channels: properties of the control channels
        TOL: time durations equal or smaller than this are considered zero (in seconds)
    """
    # pylint: disable=too-many-locals,protected-access,too-many-branches,too-many-statements
    if not B:
        return  # shortcut for empty schedules
    # add missing channels
    A.add_channels(tuple(B))

    def find_nonsolid_depth(instructions: Iterable[Instruction]) -> tuple[float, float]:
        """Returns the free and blocking nonsolid depths of the given Instruction sequence."""
        free = 0.0
        blocking = 0.0
        block_seen = False
        for inst in instructions:
            if block_seen:
                if isinstance(inst, NONSOLID):
                    # even Nothing becomes blocking depth after the first blocking instruction is encountered
                    blocking += inst.duration
                else:
                    break
            else:
                if isinstance(inst, Nothing):
                    free += inst.duration
                elif isinstance(inst, BLOCKING):
                    block_seen = True
                    blocking += inst.duration
                else:
                    break  # all other Instructions are solid
        else:
            # ran out of instructions, infinite depth
            if block_seen:
                blocking = float("inf")
            else:
                free = float("inf")
        return free, blocking

    def merge_overlap(iA: Optional[Instruction], iB: Optional[Instruction], duration: float) -> Instruction:
        """Instruction resulting from the overlap of two nonsolid Instructions.

        Nonsolid instructions are all Wait-like, and can thus be chopped up into multiple
        pieces without it changing anything. This function must never be called on solid instructions.
        """
        # make sure the channel can handle the duration
        if duration > TOL:
            _ = channels[ch].duration_to_int_samples(duration, message=f"{ch}: {iA} + {iB}: overlap duration")

        if isinstance(iA, Nothing) and isinstance(iB, Nothing):
            return Nothing(duration=duration)  # type: ignore
        return Block(duration=duration)  # type: ignore

    def find_start_time() -> float:
        """Find the earliest possible start time for schedule B."""
        start_times = {}
        for ch, segment_B in B.items():
            segment_A = A[ch]  # corresponding segment in A
            free_A, blocking_A = find_nonsolid_depth(reversed(segment_A))
            free_B, blocking_B = find_nonsolid_depth(segment_B)
            start_times[ch] = segment_A.duration - free_A - free_B - min(blocking_A, blocking_B)
        # time, relative to the start of A, where B will start
        return max(0.0, *start_times.values())

    def is_solid(inst: Instruction) -> bool:
        return not isinstance(inst, NONSOLID)

    B_start = find_start_time()
    # add the instructions from B to A
    for ch, segment_B in B.items():
        segment_A = A[ch]
        pointer_B = SegmentPointer(segment_B._instructions, 0, TOL)
        # distance between end of A and start of B
        distance = B_start - segment_A.duration
        # This is how long the merged segment_A will be.
        # We bypass the duration count system in segment_A for efficiency.
        segment_A._duration = max(B_start + segment_B.duration, segment_A.duration)

        if distance >= -TOL:
            if distance > TOL:
                # gap, add a Nothing between A and B
                segment_A._instructions.append(Nothing(duration=distance))  # type: ignore
            # add B
            segment_A._instructions.extend(pointer_B.tail())
        else:
            # there is overlap between A and B
            pointer_A = SegmentPointer(segment_A._instructions, len(segment_A), TOL)
            # find the instruction in A on which the overlap with B starts
            pointer_A.rewind(-distance)
            # pointer_A only needs to know about the overlapping tail part, which should be short in comparison to
            # the whole segment_A. segment_A is truncated to the non-overlapping part.
            pointer_A.cut_tail()

            # add partially non-overlapping item of A
            A_ran_out = False
            B_ran_out = False
            if pointer_A.frac > TOL:
                iA = pointer_A.get()
                if is_solid(iA):
                    segment_A._instructions.append(iA)
                    B_ran_out = pointer_B.fastforward(pointer_A.remainder)
                    A_ran_out = pointer_A.next()
                else:
                    segment_A._instructions.append(merge_overlap(iA, None, pointer_A.frac))

            # add overlapping items from A and B until either is exhausted
            while not A_ran_out and not B_ran_out:
                iA = pointer_A.get()
                iB = pointer_B.get()
                if is_solid(iA):
                    # solid instructions are never split, frac is always 0, and can never overlap
                    segment_A._instructions.append(iA)
                    B_ran_out = pointer_B.fastforward(pointer_A.remainder)
                    A_ran_out = pointer_A.next()
                elif is_solid(iB):
                    segment_A._instructions.append(iB)
                    A_ran_out = pointer_A.fastforward(pointer_B.remainder)
                    B_ran_out = pointer_B.next()
                else:
                    step = min(pointer_A.remainder, pointer_B.remainder)
                    segment_A._instructions.append(merge_overlap(iA, iB, step))
                    A_ran_out = pointer_A.fastforward(step)
                    B_ran_out = pointer_B.fastforward(step)

            # A, B or both ran out
            if not B_ran_out:
                # some B remains, add partially non-overlapping item of B, if any
                if pointer_B.frac > TOL:
                    iB = pointer_B.get()
                    # iB must be nonsolid, since solid instructions are not split
                    segment_A._instructions.append(merge_overlap(None, iB, pointer_B.remainder))
                    pointer_B.next()
                # add rest of B, if any
                segment_A._instructions.extend(pointer_B.tail())
            elif not A_ran_out:
                # some A remains, add partially non-overlapping item of A, if any
                if pointer_A.frac > TOL:
                    iA = pointer_A.get()
                    # iA must be nonsolid, since solid instructions are not split
                    segment_A._instructions.append(merge_overlap(iA, None, pointer_A.remainder))
                    pointer_A.next()
                # add rest of A, if any
                segment_A._instructions.extend(pointer_A.tail())


def extend_hard_boundary(
    schedule: Schedule,
    child_schedule: Schedule,
    child_components: set[str],
    neighborhood_components: set[str],
    component_durations: dict[str, int],
    is_alap: bool,
) -> None:
    """Merge two Schedules together such that the timebox boundary is respected.

    This scheduling algorithm treats the Schedules as hard, rectangular boxes where any ragged edges
    will be padded with Waits, and the boxes are not allowed to overlap.

    The algorithm is as follows:

    1. When adding ``child_schedule`` to ``schedule``, the longest channel in ``schedule`` that overlaps with the
    channels present in child determines the earliest possible starting time for the ``child_schedule``, and all other
    channels in ``schedule`` are padded with ``Wait`` to the aforementioned max length.

    2. An occupied channel in ``schedule`` will always occupy all channels of the corresponding component (qubit,
    coupler, ...). This is handled by keeping track of occupied durations for each component (no unnecessary padding
    is added to channels which do not have an actual physical pulse).

    3. After the schedules are combined, all the common channels of ``schedule`` and ``child_schedule`` are blocked
    up to their common maximum length.

    This algorithm should not be used with variable sampling rates in the schedule channels. In that case, use
    :func:`extend_hard_boundary_in_seconds` instead.

    Args:
        schedule: Schedule that should be extended with ``child_schedule``. Modified in place.
        child_schedule: Child schedule to be added.
        child_components: Components (qubits, couplers, computational_resonators) that have at least
            one channel in ``child_schedule``.
        neighborhood_components: QPU components neighboring the ``child_components`` that should
            additionally be blocked in the scheduling.
        component_durations: Blocked durations for each component used by ``schedule``.
            These act as the earliest starting points for a new segment added to any of the channels
            of the component, but will also block the component even if it has no occupied channels
            in the schedule yet or ever (e.g. a computational resonator).
            Modified in place.
        is_alap: Whether the scheduling strategy is ALAP (As Late As Possible).
    """
    # add the child's channels to the schedule if they aren't there yet
    child_channels = child_schedule.channels()
    schedule.add_channels(child_channels)

    # the child schedule can start once all the components it uses are free
    child_start: int = max((T for c, T in component_durations.items() if c in child_components), default=0)
    child_duration: int = 0
    for channel in child_channels:
        # pad the corresponding parent segment with Wait
        segment = schedule[channel]
        if segment.duration < child_start:
            segment.append(Wait(child_start - segment.duration))
        # add the child segment's instructions to parent segment
        child_segment = child_schedule[channel]
        child_instructions = reversed(child_segment) if is_alap else child_segment
        segment._instructions.extend(child_instructions)  # pylint: disable=protected-access
        segment._duration += child_segment.duration  # pylint: disable=protected-access
        # compute the child schedule duration
        if child_segment.duration > child_duration:
            child_duration = child_segment.duration

    # update the blocked durations
    for component in neighborhood_components:
        component_durations[component] = child_start + child_duration
    # the schedule's total duration is not known anymore after adding child
    schedule._duration = None


def extend_hard_boundary_in_seconds(
    schedule: Schedule,
    child_schedule: Schedule,
    child_components: set[str],
    neighborhood_components: set[str],
    component_durations: dict[str, float],
    is_alap: bool,
    channel_properties: dict[str, ChannelProperties],
) -> None:
    """The same as ``extend_hard_boundary``, but the scheduling is done in seconds.

    Used when the probe channel sampling rate differs from the other channels' rate.
    The incoming schedules measure Instruction durations in samples, but ``component_durations``
    is in seconds.

    Args:
        schedule: Schedule that should be extended with ``child_schedule``. Modified in place.
        child_schedule: Child schedule to be added.
        child_components: Components (qubits, couplers, computational_resonators) that have at least
            one channel in ``child_schedule``.
        neighborhood_components: Components neighboring the ``child_components`` that should
            additionally be blocked in the scheduling.
        component_durations: Blocked durations for each component in ``schedule``.
            These act as the earliest starting points for new segment added to any of the channels
            of a given component, but will also block the component even if it has no occupied channels
            in the schedule yet or ever (e.g. a computational resonator).
            The durations are in seconds. Modified in place.
        is_alap: Whether the scheduling strategy is ALAP (As Late As Possible).
        channel_properties: Mapping from channel name to its properties (e.g. the sample rates
            and granularities).
    """
    # add the child's channels to the schedule if they aren't there yet
    child_channels = child_schedule.channels()
    schedule.add_channels(child_channels)

    # the child schedule can start once all the components it uses are free
    child_start = max((T for c, T in component_durations.items() if c in child_components), default=0.0)
    child_duration: float = 0.0
    for channel in child_channels:
        # pad the corresponding parent segment with Wait
        segment = schedule[channel]
        channel_props = channel_properties[channel]
        segment_duration_in_seconds = channel_props.duration_to_seconds(segment.duration)
        extra_offset = 0.0
        # TOLERANCE accounts for floating point errors etc.
        if segment_duration_in_seconds + TOLERANCE < child_start:
            # wait duration that needs to be added such that the TOLERANCE is accounted for
            wait_duration_in_seconds = round(child_start - segment_duration_in_seconds, TOLERANCE_DECIMALS)
            # round the wait duration up to the closest value that can be handled by the channel
            wait_duration_rounded = channel_props.round_duration_to_granularity(
                wait_duration_in_seconds, round_up=True, force_min_duration=True
            )
            wait_duration_in_samples = channel_props.duration_to_int_samples(wait_duration_rounded)
            segment.append(Wait(wait_duration_in_samples))
            # extra time offset that was added due to the variable sampling rate
            extra_offset = wait_duration_rounded - wait_duration_in_seconds

        # add the child segment's instructions to parent segment
        child_segment = child_schedule[channel]
        child_instructions = reversed(child_segment) if is_alap else child_segment
        segment._instructions.extend(child_instructions)  # pylint: disable=protected-access
        segment._duration += child_segment.duration  # pylint: disable=protected-access

        # the increase in duration added to channel includes the extra_offset caused by variable sampling rates
        child_segment_duration_in_seconds = channel_props.duration_to_seconds(child_segment.duration) + extra_offset
        if child_segment_duration_in_seconds > child_duration:
            child_duration = child_segment_duration_in_seconds

    for component in neighborhood_components:
        component_durations[component] = child_start + child_duration
    schedule._duration = None  # the schedule's total duration is not known anymore after adding child



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Waveform definitions.

This module defines some waveforms that don't have special serialization, and reimports
waveforms that do from :mod:`iqm.models.playlist.waveforms`.
See the link for documentation of waveforms that don't appear here.
"""
# pylint: disable=no-name-in-module, unused-import
from __future__ import annotations

from dataclasses import dataclass
import logging

from iqm.models.playlist.waveforms import (
    CanonicalWaveform,
    Constant,
    CosineRiseFall,
    Gaussian,
    GaussianDerivative,
    GaussianSmoothedSquare,
    Samples,
    TruncatedGaussian,
    TruncatedGaussianDerivative,
    TruncatedGaussianSmoothedSquare,
    Waveform,
)
import numpy as np
from scipy.integrate import romb
from scipy.interpolate import interp1d
import scipy.signal as ss

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class CosineRiseFallDerivative(Waveform):
    r"""Derivative of a waveform that has a sinusoidal rise and fall, and a constant part in between.

    .. math::
        f(t) =
        \begin{cases}
          -\sin(\pi (t - c + p / 2) / r) & t - (c - p / 2) \in [-r, 0]\\
          -\sin(\pi (t - c - p / 2) / r) & t - (c + p / 2) \in [0, r]\\
          0 & \text{otherwise}
        \end{cases}

    where :math:`c` is :attr:`center_offset`, :math:`r` is :attr:`rise_time`, and :math:`p` is the plateau width,
    calculated via :math:`p :=` :attr:`full_width` - 2 * :attr:`rise_time`.

    Its values are in :math:`[-1, 1]`.


    Args:
        full_width: Duration of the support of the pulse, >= 2 * :attr:`rise_time`.
        rise_time: Duration of the sinusoidal rise (and fall) part of the waveform, >= 0.
        center_offset: The waveform is centered at this offset from the midpoint of the sampling window.
    """

    full_width: float
    rise_time: float
    center_offset: float = 0.0

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        offset_coords = sample_coords - self.center_offset
        fw = np.abs(self.full_width)
        rt = np.abs(self.rise_time)
        pw = fw - 2 * rt

        if pw < 0:
            logging.getLogger(__name__).warning(
                "Since full width is smaller than twice the rise time, the derivative of the"
                "CosineRiseFall waveform jumps at the symmetry point of the waveform."
            )
            waveform = np.piecewise(
                offset_coords,
                [
                    offset_coords <= -fw / 2,
                    offset_coords > -fw / 2,
                    offset_coords > 0,
                    offset_coords >= fw / 2,
                ],
                [
                    0,
                    lambda oc: -np.sin(np.pi / rt * (oc + pw / 2)),
                    lambda oc: -np.sin(np.pi / rt * (oc - pw / 2)),
                    0,
                ],
            )
        else:
            waveform = np.piecewise(
                offset_coords,
                [
                    offset_coords <= -pw / 2 - rt,
                    offset_coords > -pw / 2 - rt,
                    offset_coords >= -pw / 2,
                    offset_coords > pw / 2,
                    offset_coords >= pw / 2 + rt,
                ],
                [
                    0,
                    lambda oc: -np.sin(np.pi / rt * (oc + pw / 2)),
                    0,
                    lambda oc: -np.sin(np.pi / rt * (oc - pw / 2)),
                    0,
                ],
            )
        return waveform


@dataclass(frozen=True)
class Cosine(Waveform):
    r"""Periodic sinusoidal waveform which defaults to cosine.

    The use case for this waveform is to do manual modulation of other waveforms.

    .. math::
        f(t) = \cos(2\pi \: f \: t + \phi)

    where :math:`f` is the frequency, and :math:`\phi` the phase of the wave.

    Args:
        frequency: frequency of the wave, in units of inverse sampling window duration
        phase: phase of the wave, in radians
    """

    frequency: float
    phase: float = 0.0

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        return np.cos(2 * np.pi * self.frequency * sample_coords + self.phase)

    @staticmethod
    def non_timelike_attributes() -> dict[str, str]:
        return {"frequency": "Hz"}


@dataclass(frozen=True)
class Slepian(Waveform):
    r"""Slepian waveform, which minimizes non-adiabatic errors during a gate.

    It is assumed that the user has done the minimization in a prior step, such that the optimal :math:`\lambda_n` for
    a specific length :math:`\tau_\text{pulse}` (in the accelerated frame) is known. This class then reconstructs the
    waveform with the following steps:

    1.  Calculate :math:`\theta(\tau)` (Slepian in the accelerated frame and in :math:`\theta` space)
    2.  Calculate :math:`t(\tau)` (mapping time in the accelerated frame to time in the lab frame)
    3.  Interpolate :math:`\theta(t)` (Slepian in the lab frame and in :math:`\theta` space)
    4.  Calculate :math:`f(t)` (Slepian in the lab frame and in frequency space)
    5.  Calculate :math:`V(t)` (Slepian in the lab frame and in voltage space)

    Since the waveform is normalized, any voltage pre-factor should go into the pulse amplitude.

    The user is advised to look up :cite:`Martinis2014` for further details, since the derivation is mathematically
    heavy.

    Args:
        full_width: Duration of the support of the waveform.
        lambda_1: First coefficient of Slepian waveform.
        lambda_2: Second coefficient of Slepian waveform.
        frequency_initial_normalized: Initial frequency of the pulsed component (usually coupler),
            normalized by the maximum frequency of the pulsed component.
        frequency_to_minimize_normalized: Frequency of the static component (usually qubit) which to
            minimize the leakage from/to, normalized by the maximum frequency of the pulsed component.
        coupling_strength_normalized: Coupling strength between pulsed component and static component,
            normalized by the maximum frequency of the pulsed component.
        center_offset: The waveform is centered at this offset from the midpoint of the sampling window.
        squid_asymmetry: Squid asymmetry.

    """
    # pylint: disable=too-many-instance-attributes

    full_width: float
    lambda_1: float
    lambda_2: float
    frequency_initial_normalized: float
    frequency_to_minimize_normalized: float
    coupling_strength_normalized: float
    center_offset: float = 0
    squid_asymmetry: float = 0

    @staticmethod
    def non_timelike_attributes() -> dict[str, str]:
        return {
            "lambda_1": "",
            "lambda_2": "",
            "frequency_initial_normalized": "",
            "frequency_to_minimize_normalized": "",
            "coupling_strength_normalized": "",
            "squid_asymmetry": "",
        }

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        fw = np.abs(self.full_width)
        offset_coords = sample_coords - self.center_offset
        return np.piecewise(
            offset_coords,
            [(offset_coords <= -fw / 2) | (offset_coords >= fw / 2)],
            [0, lambda coords: self._sample_normalized_slepian(coords / fw + 0.5)],
        )

    def _sample_normalized_slepian(self, coords: np.ndarray) -> np.ndarray:
        """Calculate the samples of the actual Slepian waveform.

        Args:
            coords: normalized sample coordinates, in :math:`[0, 1]`
        Returns:
            samples of the Slepian waveform
        """

        def freq_to_voltage(f_normalized: float | np.ndarray, d: float) -> float | np.ndarray:
            """Map frequencies to normalized voltages.

            Here, we use the formula from `exa.core.calibration_tools.qubit_utils.simple_transmon_ge_frequency`, and
            solve the equation for phi.
            Note that we don't need the voltage period, since it just scales the amplitude, which is anyways normalized
            for every pulse to 1. Furthermore, we don't need the voltage offset, since we are only interested in the
            frequency difference from the initial frequency.
            Args:
                f_normalized: Frequency to be converted to voltage normalized over the sweet spot frequency.
                d: Squid asymmetry.
            Returns:
                Voltage corresponding to the input frequency.

            """
            return -np.arccos(np.sqrt((f_normalized**4 - d**2) / (1 - d**2)))

        # build an interpolating mapping from t to theta
        tau_array = np.linspace(0, 1, 51)  # tau is normalized to [0, 1]
        interp_t_to_theta = interp1d(
            self._t_tau(tau_array), self._theta_tau(tau_array), kind="linear", bounds_error=True
        )

        # the waveform is scaled to the required duration
        t = coords * self._t_tau(1)
        frequency = self.frequency_to_minimize_normalized + 2 * self.coupling_strength_normalized / np.tan(
            interp_t_to_theta(t)
        )
        slepian_samples = freq_to_voltage(frequency, self.squid_asymmetry) - freq_to_voltage(
            self.frequency_initial_normalized, self.squid_asymmetry
        )
        return np.abs(slepian_samples / np.max(np.abs(slepian_samples)))

    def _t_tau(self, tau: float | np.ndarray, num_samples: int = 2**7 + 1) -> np.ndarray:
        r"""Convert time in the accelerated frame to the lab frame (real time).

        Since the conversion is defined via an integral, finding an analytical solution was not possible. Therefore, we
        integrate the expression numerically with :func:`scipy.integrate.romb`.

        Args:
            tau: Time in the accelerated frame.
            num_samples: Number of samples for the numerical integration. Must be of the form 2 ** k + 1.

        Returns:
            ``tau`` converted to the lab frame
        """
        tau = np.atleast_1d(tau)
        t_array = []
        for tau_single in tau:
            if tau_single == 0:
                t_array.append(0)
            else:
                tau_samples = np.linspace(0, tau_single, num_samples)
                integrand = np.sin(self._theta_tau(tau_samples))
                t = romb(integrand, float(tau_samples[1] - tau_samples[0]))
                t_array.append(t)

        return np.abs(np.array(t_array))

    def _theta_tau(self, tau: np.ndarray) -> np.ndarray:
        r"""Parametrization of the Slepian waveform in the accelerated frame.

        The Slepian waveform is parametrized using Fourier base functions, where we only take the cosine terms into
        account (:cite:`Martinis2014` has shown that this is a reasonable assumption).
        Here, :math:`\lambda_n` is the amplitude of the :math:`n`-th term of the Fourier base function.
        It is usually sufficient to take only up to second order terms into account, i.e. only :math:`\lambda_1` and
        :math:`\lambda_2` are non-zero.

        Args:
            tau: Time in the accelerated frame, normalized to [0, 1] with tau_pulse.

        Returns:
            Slepian waveform in the theta space and accelerated frame.
        """
        lambdas = np.array([self.lambda_1, self.lambda_2])
        theta_i = np.arctan(
            2
            * self.coupling_strength_normalized
            / (self.frequency_initial_normalized - self.frequency_to_minimize_normalized)
        )
        return theta_i + np.sum(
            [lambda_i * (1 - np.cos(2 * np.pi * (i + 1) * tau)) for i, lambda_i in enumerate(lambdas)],
            axis=0,
        )


@dataclass(frozen=True)
class Chirp(Waveform):
    r"""Linear chirp, defined as

     .. math:: f(t) = A \: \omega[\alpha, N] \: \cos(2\pi \int (f_{0} + (f_{1} - f_{0}) t) \: \mathrm{d}t),

    where :math:`\omega[\alpha, N]` is a cosine-tapered window. For :math:`\alpha = 1` it becomes rectangular,
    and for :math:`\alpha = 0` it becomes a Hann (or raised cosine) window.

    The chirp pulse is valued inside the Nyquist zone, such that :math:`f_{0}` and :math:`f_{1}` are constrained
    in the range :math:`[-0.5, 0.5]`.

    Args:
        freq_start: Initial frequency of the chirp waveform in the Nyquist zone.
        freq_stop: Final frequency of the chirp waveform in the Nyquist zone.
        alpha: Alpha parameter of the cosine-tapered window. Defaults to 0.05.
        phase: Phase of the waveform. Defaults to 0
    """
    freq_start: float
    freq_stop: float
    alpha: float = 0.05
    phase: float = 0

    def _sample(self, sample_coords):

        chirpfreq = np.linspace(self.freq_start, self.freq_stop, len(sample_coords))
        chirpphase = 2 * np.pi * np.cumsum(chirpfreq) + self.phase
        wave = np.exp(1j * chirpphase) * ss.windows.tukey(len(sample_coords), self.alpha)
        return wave.real

    @staticmethod
    def non_timelike_attributes() -> dict[str, str]:
        return {
            "alpha": "",
            "phase": "",
            "freq_start": "",
            "freq_stop": "",
        }


@dataclass(frozen=True)
class ChirpImag(Chirp):
    """Imaginary part of the linear chirp, which sets the phase to $-\\pi/2$.

    Attributes:
        phase: Phase of the pulse. Defaults to $\\pi/2$
    """

    phase: float = -np.pi / 2


@dataclass(frozen=True)
class ModulatedCosineRiseFall(Waveform):
    r"""Modulated Cosine Rise Fall waveform.

    This waveform takes the waveform :class:`CosineRiseFall` and modulates it with a cosine signal
    which then has parameters :attr:`frequency` and :attr:`phase`, additional to the parameters :attr:`full_width`,
    :attr:`rise_time`, and :attr:`center_offset`, see description of :class:`TruncatedGaussianSmoothedSquare` for
    further details.


    Args:
        full_width: Full width of the pulse, >= 2 * :attr:`rise_time`.
        rise_time: Duration of the sinusoidal rise (and fall) part of the waveform, >= 0.
        modulation_frequency: Modulation frequency.
        phase: Phase of the modulation.
        center_offset: The waveform is centered around this sampling window coordinate.
            If zero, the pulse is placed in the middle of the sampling window.
    """

    full_width: float
    rise_time: float
    modulation_frequency: float = 0.0
    phase: float = 0
    center_offset: float = 0.0

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        offset_coords = sample_coords - self.center_offset
        fw = np.abs(self.full_width)
        rt = np.abs(self.rise_time)
        pw = fw - 2 * rt

        if pw < 0:
            logging.getLogger(__name__).warning(
                "Since the full width is smaller than twice the rise time, the CosineRiseFall pulse does not reach its"
                "top.",
            )
            envelope = np.piecewise(
                offset_coords,
                [
                    offset_coords <= -fw / 2,
                    offset_coords > -fw / 2,
                    offset_coords > 0,
                    offset_coords >= fw / 2,
                ],
                [
                    0,
                    lambda oc: 0.5 + 0.5 * np.cos(np.pi / rt * (oc + pw / 2)),
                    lambda oc: 0.5 + 0.5 * np.cos(np.pi / rt * (oc - pw / 2)),
                    0,
                ],
            )
        else:
            envelope = np.piecewise(
                offset_coords,
                [
                    offset_coords <= -pw / 2 - rt,
                    offset_coords > -pw / 2 - rt,
                    offset_coords >= -pw / 2,
                    offset_coords > pw / 2,
                    offset_coords >= pw / 2 + rt,
                ],
                [
                    0,
                    lambda oc: 0.5 + 0.5 * np.cos(np.pi / rt * (oc + pw / 2)),
                    1,
                    lambda oc: 0.5 + 0.5 * np.cos(np.pi / rt * (oc - pw / 2)),
                    0,
                ],
            )
        return envelope * np.cos(2 * np.pi * self.modulation_frequency * sample_coords + self.phase)

    @staticmethod
    def non_timelike_attributes() -> dict[str, str]:
        return {"modulation_frequency": "Hz", "phase": "rad"}



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Single-qubit SU(2) gate.

The SU(2) matrix in the computational basis is parametrized using Euler angles:


.. math::
   U(\theta, \phi, \lambda) =
    \begin{pmatrix}
    cos(\theta / 2) & -e^{i\lambda}\sin{\theta/2} \\
    e^{i\phi}\sin{\theta/2} & e^{i(\lambda+\phi)}\cos{\theta/2}
    \end{pmatrix}

where the angles :math:`\theta`, :math:`\phi` and :math:`\lambda` are in radians. They are the angles of subsequent
Z, Y and Z Euler rotations:

.. math::
    U(\theta, \phi, \lambda) = R_Z(\phi) \: R_Y(\theta) \: R_Z(\lambda)

It rotates the qubit state around an arbitrary axis on the Bloch sphere.

Some common single-qubit gates expressed as U gates:

.. math::
   X = U(\pi, -\pi/2, \pi/2)\\
   Y = U(\pi, 0, 0)\\
   Z = U(0, 0, \pi)\\
   H = U(\pi / 2, 0, \pi)\\
   S = U(0, \pi / 4, \pi / 4)\\
   T = U(0, \pi / 8, \pi / 8)

References
----------

https://openqasm.com/language/gates.html#built-in-gates
"""
from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np

from iqm.pulse.gate_implementation import CompositeGate
from iqm.pulse.playlist.schedule import Schedule
from iqm.pulse.utils import normalize_angle, phase_transformation

if TYPE_CHECKING:  # pragma: no cover
    from iqm.pulse.timebox import TimeBox


@lru_cache
def get_unitary_u(theta: float, phi: float, lam: float) -> np.ndarray:
    """Unitary for an SU(2) gate.

    See :mod:`iqm.pulse.gates.u` for the definition of the gate parameters.

    Args:
        theta: y rotation angle
        phi: z rotation angle
        lam: another z rotation angle

    Returns:
        2x2 unitary representing ``u(theta, phi, lam)``.
    """
    return np.array(
        [
            [np.cos(theta / 2), -np.exp(1j * lam) * np.sin(theta / 2)],
            [np.exp(1j * phi) * np.sin(theta / 2), np.exp(1j * (lam + phi)) * np.cos(theta / 2)],
        ]
    )


class UGate(CompositeGate):
    r"""SU(2) gate implemented using PRX.

    Assumes the chosen PRX implementation uses resonant driving, and that the virtual RZ technique can be used.
    """

    registered_gates = ["prx"]

    def _call(self, theta: float, phi: float = 0.0, lam: float = 0.0) -> TimeBox:  # type: ignore[override]
        r"""Convert pulses into timebox, via Euler decomposition.

        .. math::
            U(\theta, \phi, \lambda) = R_Z(\phi) \cdot R_Y(\theta) \cdot R_Z(\lam)
        """
        # TODO we directly modify the PRX timebox contents here which makes a lot of assumptions about
        # the PRX implementation. This isn't safe in general, can we find a better solution?

        prx_gate = self.build("prx", self.locus)
        pulse_train = prx_gate(theta, np.pi / 2).atom[prx_gate.channel]  # RY pulse

        # Check if the pulse train have one or several pulses.
        if len(pulse_train) == 1:
            # Assumes the PRX consists of a single IQPulse.
            pulse = pulse_train[0]
            new_phase, new_phase_increment = phase_transformation(lam, phi)
            new_pulse = pulse.copy(
                scale_i=pulse.scale_i,
                scale_q=pulse.scale_q,
                phase=normalize_angle(pulse.phase + new_phase),
                phase_increment=normalize_angle(pulse.phase_increment + new_phase_increment),
            )
            timebox = self.to_timebox(Schedule({prx_gate.channel: [new_pulse]}))

        else:
            # Assumes the PRX pulse train begins and ends with IQPulses.
            # Only the first and last pulse need to be changed to implement the RZs.
            pulse_a = pulse_train[0]
            pulse_b = pulse_train[-1]
            _lam_phase, _lam_phase_increment = phase_transformation(lam, 0)
            _phi_phase, _phi_phase_increment = phase_transformation(0, phi)
            new_pulse_a = pulse_a.copy(
                scale_i=pulse_a.scale_i,
                scale_q=pulse_a.scale_q,
                phase=normalize_angle(pulse_a.phase + _lam_phase),
                phase_increment=normalize_angle(pulse_a.phase_increment + _lam_phase_increment),
            )
            new_pulse_b = pulse_b.copy(
                scale_i=pulse_b.scale_i,
                scale_q=pulse_b.scale_q,
                phase=normalize_angle(pulse_b.phase + _phi_phase),
                phase_increment=normalize_angle(pulse_b.phase_increment + _phi_phase_increment),
            )
            other_pulses = [pulse.copy() for pulse in pulse_train[1:-1]]
            new_pulses = [new_pulse_a] + other_pulses + [new_pulse_b]
            timebox = self.to_timebox(Schedule({prx_gate.channel: new_pulses}))

        timebox.neighborhood_components[0] = set(self.locus)
        return timebox



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classically controlled gates."""

from exa.common.data.parameter import CollectionType, Parameter
import numpy as np

from iqm.pulse.gate_implementation import CompositeGate
from iqm.pulse.gates.measure import FEEDBACK_KEY
from iqm.pulse.gates.prx import PRX_SinglePulse_GateImplementation
from iqm.pulse.playlist.instructions import Block, ConditionalInstruction, Wait
from iqm.pulse.playlist.schedule import Schedule
from iqm.pulse.timebox import TimeBox


class CCPRX_Composite(CompositeGate):
    """Classically controlled PRX gate.

    Applies a PRX gate conditioned on a discriminated readout result obtained in the same segment (active feedback).
    Applies a PRX gate if the result is 1, and a Wait of equal duration if the result is 0.
    Uses the default implementation of PRX underneath, so no extra calibration is needed.

    Not usable in the "legacy" readout programming mode.
    """

    parameters = {"control_delays": Parameter("", "Control delays", "s", collection_type=CollectionType.NDARRAY)}
    """``control_delays`` contains the times it takes for the classical control signal from each
    probe line (readout instrument) to become usable for the drive AWG implementing the PRX gate.
    The delays must be in the same order as the probe lines are listed in
    the ``{drive_controller}.awg.feedback_sources`` station setting.
    """

    def _call(
        self, angle: float = np.pi, phase: float = 0.0, *, feedback_qubit: str, feedback_key: str
    ) -> list[TimeBox]:
        """Two TimeBoxes that together implement the classically controlled PRX gate.

        The first Timebox is for the control signal delay, and the second has a ConditionalInstruction.
        The delay TimeBox operates only on a virtual channel and is used to block the pulse TimeBox
        until there has been enough time for the control signal to arrive.
        The delay is specified by the ``control_delays`` gate parameter.

        In normal operation, the boxes can be placed sequentially without causing unnecessary delays.
        To care of the timing yourself, simply ignore the first TimeBox.

        Args:
            angle: The PRX rotation angle (rad).
            phase: The PRX rotation phase (rad).
            feedback_qubit: The qubit that was measured to create the feedback bit.
            feedback_key: Identifies the feedback signal if ``feedback_qubit`` was measured multiple times.
                The feedback label is then ``f"{feedback_qubit}__{feedback_key}"``.
                TODO: currently the HW does not support multiple feedback keys per drive channel, so this argument has
                no effect. The conditional prx will always listen feedback from the label
                ``f"{feedback_qubit}__{FEEDBACK_KEY}"``. When the HW is improved, the actual key the user inputs
                should be used.

        Returns:
            A TimeBox for the signal delay, and a TimeBox with a ConditionalInstruction inside.
        """
        # pylint: disable=unused-argument
        qubit = self.locus[0]
        awg_name = self.builder.get_drive_channel(qubit)

        prx_gate: PRX_SinglePulse_GateImplementation = self.build("prx", self.locus)
        # FIXME assumes PRX gates only use this one implementation as the default,
        # with just a drive channel and a single IQPulse.
        pulse = prx_gate(angle, phase).atom[prx_gate.channel][0]
        wait = Wait(pulse.duration)  # idling, can be replaced with a DD sequence later on

        # TODO: use the actual inputted label when the HW supports many labels per drive channel
        default_label = f"{feedback_qubit}__{FEEDBACK_KEY}"
        pulse_instruction = ConditionalInstruction(
            duration=pulse.duration, condition=default_label, outcomes=(wait, pulse)
        )
        delays = self.builder.get_calibration(self.parent.name, self.name, self.locus)["control_delays"]
        if len(delays) == 0:
            raise ValueError(f"'control_delays' for '{self.name}' on {qubit} is empty (not calibrated).")

        possible_sources = [c for c in self.builder.get_virtual_feedback_channels(qubit) if awg_name in c]
        if len(delays) != len(possible_sources):
            raise ValueError(
                f"Not the correct amount of calibration values for 'control_delays'. Need {len(possible_sources)}"
                f"values, got {delays}."
            )
        virtual_channel_name = self.builder.get_virtual_feedback_channel_for(awg_name, feedback_qubit)
        delay = delays[possible_sources.index(virtual_channel_name)]
        virtual_channel = self.builder.channels[virtual_channel_name]
        delay_samples = virtual_channel.duration_to_int_samples(
            virtual_channel.round_duration_to_granularity(delay, round_up=True), check_min_samples=False
        )
        delay_box = TimeBox.atomic(
            Schedule({virtual_channel_name: [Block(delay_samples)]}),
            locus_components=[],
            label=f"Feedback signal delay for {qubit}",
        )
        delay_box.neighborhood_components = {0: {virtual_channel_name}}
        cond = TimeBox.atomic(
            Schedule({virtual_channel_name: [Block(0)], prx_gate.channel: [pulse_instruction]}),
            locus_components=[qubit],
            label=f"Conditional PRX for {qubit}",
        )
        cond.neighborhood_components = {0: {virtual_channel_name, qubit}}
        return [
            delay_box,
            cond,
        ]


class CCPRX_Composite_DRAGCosineRiseFall(CCPRX_Composite):
    """Conditional drag_crf pulse."""

    default_implementations = {"prx": "drag_crf"}


class CCPRX_Composite_DRAGGaussian(CCPRX_Composite):
    """Conditional drag_gaussian pulse."""

    default_implementations = {"prx": "drag_gaussian"}



#  ********************************************************************************
#
# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Function for visualising playlists."""
import base64
from collections.abc import Sequence
from dataclasses import asdict
import html
import io
import json
from pathlib import Path

from iqm.models.playlist.waveforms import Waveform
from jinja2 import Environment, FileSystemLoader
import numpy as np

from iqm.pulse.playlist.playlist import Playlist


def _get_waveform(wave: Waveform, scale: float, wave_q: Waveform | None = None, scale_q: float | None = None) -> str:
    # pylint: disable=import-outside-toplevel, import-error, consider-using-f-string
    import matplotlib.pyplot as plt

    def fig_to_base64(fig):
        img = io.BytesIO()
        fig.savefig(img, format="png", bbox_inches="tight")
        img.seek(0)

        return base64.b64encode(img.getvalue())

    scaled_wave = wave.sample() * scale
    if wave_q:
        scaled_wave_q = wave_q.sample() * scale_q
    html_text = ""
    with plt.ioff():
        fig, ax = plt.subplots(1, 1)
        ax.set_ylim([-1, 1])
        ax.plot(scaled_wave, marker=".")
        if wave_q:
            ax.plot(scaled_wave_q, marker=".")
        encoded = fig_to_base64(fig)
        my_html = '<img class="waveform-image" src="data:image/png;base64, {} ">'.format(encoded.decode("utf-8"))
        html_text = my_html
        plt.close(fig)
    return html_text


def _playlist_as_a_dict(playlist: Playlist, segment_indices: Sequence[int]) -> dict:
    # pylint: disable=too-many-locals,too-many-branches
    playlists: list = []
    waveform_dict: dict[str, dict] = {}
    for channel in playlist.channel_descriptions:
        waveform_dict[channel] = {}

    for idx in segment_indices:
        segment = playlist.segments[idx]
        schedule_dict: dict[str, dict] = {}
        for channel, instruction_list in segment.instructions.items():
            schedule_dict[channel] = {}
            channel_desc = playlist.channel_descriptions.get(channel, None)
            instructions: list[dict] = []
            for instruction_idx in instruction_list:
                instruction = channel_desc.instruction_table[instruction_idx]
                instruction_dict = {}
                instruction_dict["name"] = instruction.operation.__class__.__name__
                params = asdict(instruction.operation)
                if instruction.operation.__class__.__name__ == "RealPulse":
                    instruction_dict["wave_img_idx"] = instruction_idx
                    if str(instruction_idx) not in waveform_dict[channel]:
                        waveform_dict[channel][str(instruction_idx)] = _get_waveform(
                            instruction.operation.wave, instruction.operation.scale
                        )
                elif instruction.operation.__class__.__name__ == "IQPulse":
                    instruction_dict["wave_img_idx"] = instruction_idx
                    if str(instruction_idx) not in waveform_dict[channel]:
                        waveform_dict[channel][str(instruction_idx)] = _get_waveform(
                            instruction.operation.wave_i,
                            instruction.operation.scale_i,
                            instruction.operation.wave_q,
                            instruction.operation.scale_q,
                        )
                elif instruction.operation.__class__.__name__ == "ReadoutTrigger":
                    if instruction.operation.probe_pulse.operation.__class__.__name__ == "MultiplexedIQPulse":
                        params = {
                            "probe_pulse": instruction.operation.probe_pulse.operation.__class__.__name__,
                            "entries": [
                                json.dumps(
                                    {"name": entry[0].operation.__class__.__name__, **asdict(entry[0].operation)},
                                    indent=2,
                                )
                                for entry in instruction.operation.probe_pulse.operation.entries
                            ],
                        }
                params["duration"] = round(
                    instruction.duration_samples / channel_desc.channel_config.sampling_rate * 1e9, 3
                )

                instruction_dict["params"] = params
                instructions.append(instruction_dict)
            schedule_dict[channel]["instructions"] = instructions

        # squeeze start and end waits
        for squeeze_idx in (0, -1):
            common_wait = np.inf
            for channel, content in schedule_dict.items():
                if "instructions" in content and content["instructions"][squeeze_idx]["name"] == "Wait":
                    common_wait = min(common_wait, content["instructions"][squeeze_idx]["params"]["duration"])
                else:
                    common_wait = 0
                    break
            if common_wait:
                for channel, content in schedule_dict.items():
                    content["instructions"][squeeze_idx]["params"]["duration"] -= common_wait
                    qualif = "end" if squeeze_idx == -1 else "start"
                    common_wait_instr = {
                        "name": f"Wait at {qualif}",
                        "params": {"duration": common_wait, "truncated_duration": 150},
                    }
                    content["instructions"].insert(squeeze_idx, common_wait_instr)
        playlists.append(schedule_dict)
    ret_dict = {"playlists": playlists, "waveforms": waveform_dict}

    return ret_dict


def inspect_playlist(playlist: Playlist, segments: Sequence[int] = (0,)) -> str:
    """Creates an HTML string from the given playlist and segments.

    The output can be viewed in a browser or in a Jupyter notebook using ``IPython.core.display.HTML``.

    Args:
        playlist: The Playlist to be visualised
        segments: Indices of the Playlist segments to inspect.

    Returns:
       The generated raw HTML string.
    """
    path = Path(__file__).parent / "templates"
    file_loader = FileSystemLoader(path)
    env = Environment(loader=file_loader)
    template_file = "playlist_inspection.jinja2"
    for value in segments:
        try:
            playlist.segments[value]
        except IndexError as exc:
            end_range = ""
            if len(playlist.segments) > 1:
                end_range = f"-{len(playlist.segments) - 1}"
            raise IndexError(f"Index '{value}' not in range of segments 0{end_range}") from exc
    template = env.get_template(template_file, globals={"round": round})
    json_format = _playlist_as_a_dict(playlist=playlist, segment_indices=segments)
    vis_js_style = Path(path / "static" / "vis-timeline-graph2d.min.css", encoding="utf-8").read_text(encoding="utf-8")
    vis_js_script = Path(path / "static" / "vis-timeline-graph2d.min.js", encoding="utf-8").read_text(encoding="utf-8")
    moment_js_script = Path(path / "static" / "moment.min.js", encoding="utf-8").read_text(encoding="utf-8")

    html_text = template.render(
        jsonobj=json_format,
        VisJsScript=vis_js_script,
        VisJsStyle=vis_js_style,
        MomentJS=moment_js_script,
        segment_indices=list(segments),
    )
    return f"""<iframe allowfullscreen="true"
        style="background: #F4F0EA; width: 100%; height: 1000;"
        width=1000 height=600
        srcdoc="{html.escape(html_text)}">"""



#  ********************************************************************************
#
# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions."""
from __future__ import annotations

import copy
from typing import Any, Sized, get_args, get_origin

from exa.common.data.parameter import CollectionType, DataType
import numpy as np
from ruamel.yaml import YAML


def merge_dicts(A: dict, B: dict, path=(), merge_nones: bool = True) -> dict:
    """Merge two dictionaries recursively, leaving the originals unchanged.

    Args:
        A: dictionary
        B: another dictionary
        merge_nones: whether to also merge ``None`` and empty ``Sized`` values from B to A.

    Returns:
        copy of A, with the contents of B merged in (and taking precedence) recursively
    """

    def is_not_empty(val: Any) -> bool:
        if val is None:
            return False
        if isinstance(val, Sized) and len(val) == 0:
            return False
        return True

    # A and B must be left intact, make a shallow copy
    A = copy.copy(A)
    for key, vb in B.items():
        if (va := A.get(key)) is not None:
            new_path = (*path, key)
            if isinstance(va, dict):
                if isinstance(vb, dict):
                    # replace the dict with the shallow copy returned by merge_dicts
                    A[key] = merge_dicts(va, vb, new_path, merge_nones=merge_nones)
                    continue
                raise ValueError(f"Merging dict with scalar: {'.'.join(new_path)}")
            if isinstance(vb, dict):
                raise ValueError(f"Merging scalar with dict: {'.'.join(new_path)}")
        # scalar overrides scalar, or a new key is inserted
        if merge_nones or is_not_empty(vb):
            A[key] = vb
    return A  # return the shallow copy


def map_waveform_param_types(type_hint: type) -> tuple[DataType, CollectionType]:
    """Map a python typehint into EXA Parameter's `(DataType, CollectionType)` tuple.

    Args:
        type: python typehint.

    Returns:
        A `(DataType, CollectionType)` tuple

    Raises:
        ValueError: for a non-supported type.
    """
    value_error = ValueError(f"Nonsupported datatype for a waveform parameter: {type_hint}")
    if hasattr(type_hint, "__iter__") and type_hint != str:
        if type_hint == np.ndarray:
            data_type = DataType.COMPLEX  # due to np.ndarray not being generic we assume complex numbers
            collection_type = CollectionType.NDARRAY
            return (data_type, collection_type)
        if get_origin(type_hint) == list:
            collection_type = CollectionType.LIST
            type_hint = get_args(type_hint)[0]
        else:
            raise value_error
    else:
        collection_type = CollectionType.SCALAR

    if type_hint in {int, float}:
        data_type = DataType.NUMBER
    elif type_hint == str:
        data_type = DataType.STRING
    elif type_hint == complex:
        data_type = DataType.COMPLEX
    elif type_hint == bool:
        data_type = DataType.BOOLEAN
    else:
        raise value_error
    return (data_type, collection_type)


def load_yaml(path: str) -> dict[str, Any]:
    """Load a YAML file from the given path, raise error if the file can't be loaded.

    Args:
        path: path to a YAML file

    Returns:
        contents of the YAML file as Python types
    """
    with open(path, encoding="utf-8") as f:
        yaml = YAML(typ="safe")  # default, if not specified, is 'rt' (round-trip)
        data = yaml.load(f)
    return data


def normalize_angle(angle: float) -> float:
    """Normalize the given angle to (-pi, pi].

    Args:
        angle: angle to normalize (in radians)

    Returns:
        ``angle`` normalized to (-pi, pi]
    """
    half_turn = np.pi
    full_turn = 2 * half_turn
    return (angle - half_turn) % -full_turn + half_turn


def phase_transformation(psi_1: float = 0.0, psi_2: float = 0.0) -> tuple[float, float]:
    r"""Implement an arbitrary (RZ, PRX, RZ) gate sequence by modifying the parameters of the
    IQ pulse implementing the PRX.

    By commutation rules we have

    .. math::
       RZ(\psi_2) \: PRX(\theta, \phi) \: RZ(\psi_1) = PRX(\theta, \phi+\psi_2) \: RZ(\psi_1 + \psi_2).

    Hence an arbitrary (RZ, PRX, RZ) gate sequence is equivalent to (RZ, PRX) with adjusted angles.

    Use case: with resonant driving, the PRX gate can be implemented using an :class:`IQPulse` instance,
    and the preceding RZ can be handled by decrementing the local oscillator phase beforehand (something
    the IQPulse instruction can also do), which is equivalent to rotating the local computational frame
    around the z axis in the opposite direction of the required quantum state rotation.

    Args:
        psi_1: RZ angle before the PRX (in rad)
        psi_2: RZ angle after the PRX (in rad)

    Returns:
        change to the PRX phase angle (in rad),
        phase increment for the IQ pulse that implements the remaining RZ (in rad)
    """
    return psi_2, -(psi_1 + psi_2)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Barrier metaoperation.

The barrier is an n-qubit metaoperation that forces a specific temporal ordering on the quantum
operations on different sides of it (the ones preceding the barrier are always executed first).
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from iqm.pulse.gate_implementation import GateImplementation, Locus, OILCalibrationData
from iqm.pulse.playlist.instructions import Block
from iqm.pulse.playlist.schedule import Schedule

if TYPE_CHECKING:  # pragma: no cover
    from iqm.pulse.builder import ScheduleBuilder
    from iqm.pulse.quantum_ops import QuantumOp
    from iqm.pulse.timebox import TimeBox


class Barrier(GateImplementation):
    """GateImplementation for the n-qudit ``barrier`` metaoperation.

    Returns a schedule with zero-duration :class:`.Block` metainstructions.
    When this is appended to another :class:`.Schedule`,
    it causes the affected channels to be padded with :class:`.Wait` instructions to the same length,
    which in turn imposes a definite temporal order for the operations on different sides of
    the barrier (the ones preceding it are always executed first).

    .. note::
       Assumes that all instructions involve either the drive, flux or probe channels of the locus QPU components.

    Args:
        channels: channels related to the locus QPU components, to be blocked
    """

    symmetric: bool = True

    def __init__(
        self,
        parent: QuantumOp,
        name: str,
        locus: Locus,
        calibration_data: OILCalibrationData,
        builder: ScheduleBuilder,
    ):
        super().__init__(parent, name, locus, calibration_data, builder)
        channels = builder.get_control_channels(locus)
        self.schedule = Schedule({c: [Block(0)] for c in channels})

    def _call(self) -> TimeBox:
        return self.to_timebox(self.schedule)  # TODO make a copy of schedule??

    def duration_in_seconds(self) -> float:
        return 0.0



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Validation of Schedule related data."""
from __future__ import annotations

from dataclasses import dataclass

from iqm.models.playlist.channel_descriptions import ChannelDescription
from iqm.models.playlist.instructions import (
    ConditionalInstruction,
    Instruction,
    IQPulse,
    Operation,
    ReadoutTrigger,
    RealPulse,
)
import iqm.models.playlist.instructions as sc

from iqm.pulse.playlist.channel import ChannelProperties
import iqm.pulse.playlist.instructions as front
from iqm.pulse.playlist.playlist import Playlist


class PlaylistValidationError(Exception):
    """Error raised when Playlist validation fails."""


def validate_playlist_compatibility(playlist: Playlist, device_constraints: dict[str, ChannelProperties]) -> None:
    """Validates that the given playlist is compatible with the provided AWG data.

    The following requirements are validated as they are the same for all controllers.

    1. Playlist sample rate vs. Actual controller sample rate
    2. Instruction granularity with respect to the controllers granularity requirements
    3. Checks that no other than supported instructions are used.
    4. Checks that instruction length matches waveform length in IQ and RealPulse
    5. Checks that all instructions are at least the length of minimum number of samples
    6. Checks that Conditional instruction has the same duration in every child instruction

    Args:
        playlist: instructions used on each channel, as well as the channel configurations
        device_constraints: actual hardware limitations of the channels
    """
    for channel_name, channel_description in playlist.channel_descriptions.items():
        validate_channel(channel_description, device_constraints[channel_name])


def validate_channel(channel_description: ChannelDescription, device_constraints: ChannelProperties) -> None:
    """Validate a single channel."""
    channel_name = channel_description.controller_name
    # TODO remove once we use the iqm.pulse Instruction classes on both frontend and SC
    _instruction_map = {
        sc.Wait: front.Wait,
        sc.RealPulse: front.RealPulse,
        sc.IQPulse: front.IQPulse,
        sc.VirtualRZ: front.VirtualRZ,
        sc.ConditionalInstruction: front.ConditionalInstruction,
    }

    if not hasattr(channel_description.channel_config, "sampling_rate"):
        raise PlaylistValidationError("Channel configuration does not have sampling rate")

    # validate the channel config
    if channel_description.channel_config.sampling_rate != device_constraints.sample_rate:
        raise PlaylistValidationError(
            f"Sample rates do not match. "
            f"Device {channel_name} expects {device_constraints.sample_rate} "
            f"but got from playlist {channel_description.channel_config.sampling_rate}"
        )
    # validate the instructions
    allowed_names = tuple(map(lambda x: x.__name__, device_constraints.compatible_instructions))

    for instruction in channel_description.instruction_table:
        instruction_type = type(instruction.operation)
        mapped_type = _instruction_map[instruction_type]
        if mapped_type not in device_constraints.compatible_instructions:
            raise PlaylistValidationError(
                f"{channel_name}: Incompatible instruction {mapped_type.__name__} (allowed: {allowed_names})"
            )
        try:
            pass
            # TODO turn back on and remove pragma once we use the iqm.pulse Instruction classes on both frontend and SC
            # instruction.validate()
        except ValueError as ex:  # pragma: no cover
            raise PlaylistValidationError(ex) from ex

        if (instruction.duration_samples % device_constraints.instruction_duration_granularity) != 0:
            raise PlaylistValidationError(
                f"Instruction length of {instruction.duration_samples} doesn't match the "
                f"{channel_name} granularity of the device"
            )
        if instruction.duration_samples < device_constraints.instruction_duration_min:
            raise PlaylistValidationError(
                f"Instruction length of {instruction.duration_samples} is less than the minimum number "
                f"of samples {device_constraints.instruction_duration_min} for {channel_name}"
            )


@dataclass()
class AWGScheduleValidationData:
    """Controller specific validation data"""

    sampling_rate: float
    granularity: int
    min_number_of_samples: int
    compatible_instructions: tuple[type[Operation], ...]  # type: ignore[valid-type]


class AWGScheduleValidationError(Exception):
    """Error raised when schedule validation for an AWG fails."""


class InvalidInstructionError(Exception):
    """Error raised when encountering an invalid instruction."""

    def __init__(self, instruction, issue_string="unknown reason"):
        self.issue_string = issue_string
        self.instruction = instruction
        super().__init__(issue_string)

    def __str__(self):
        return f"{self.issue_string} (in {self.instruction})"


def validate_instruction_and_wf_length(instruction: Instruction):
    """Validate that instruction and waveform lengths match

    Args:
        instruction: The IQPulse or RealPulse to be validated
    """
    if isinstance(instruction.operation, RealPulse):
        if instruction.duration_samples != instruction.operation.wave.n_samples:
            raise InvalidInstructionError(instruction, "Instruction duration != waveform length")
        if abs(instruction.operation.scale) > 1.0:
            raise InvalidInstructionError(instruction, "scale not in -1..1")
    if isinstance(instruction.operation, IQPulse):
        if instruction.duration_samples != instruction.operation.wave_i.n_samples:
            raise InvalidInstructionError(instruction, "Instruction duration != waveform_i length")
        if instruction.duration_samples != instruction.operation.wave_q.n_samples:
            raise InvalidInstructionError(instruction, "Instruction duration != waveform_q length")
        if abs(instruction.operation.scale_i) > 1 or abs(instruction.operation.scale_q) > 1:
            raise InvalidInstructionError(instruction, "scale not in -1..1")


def validate_awg_and_schedule_compatibility(
    channel_description: ChannelDescription, device_constraints: AWGScheduleValidationData
):
    """Validates that the given playlist is compatible with the provided AWG data.
    The following requirements are validated as they are the same for all controllers.

    1. Playlist sampling rate vs. Actual controller sampling rate
    2. Instruction granularity with respect to the controllers granularity requirements
    3. Checks that no other than supported instructions are used.
    4. Checks that instruction length matches waveform length in IQ and RealPulse
    5. Checks that all instructions are at least the length of minimum number of samples
    6. Checks that Conditional instruction has the same duration in every child instruction

    Args:
        channel_description: Contains instructions used as well as the channel specific configuration from playlist
        device_constraints: Contains the actual hardware limitations
    """

    if channel_description is None:
        raise AWGScheduleValidationError("Controller was not found in the playlist")

    if not hasattr(channel_description.channel_config, "sampling_rate"):
        raise AWGScheduleValidationError("Channel configuration does not have sampling rate")

    if channel_description.channel_config.sampling_rate != device_constraints.sampling_rate:
        raise AWGScheduleValidationError(
            f"Sampling rates do not match. "
            f"Device {channel_description.controller_name} expects {device_constraints.sampling_rate} "
            f"but got from playlist {channel_description.channel_config.sampling_rate}"
        )
    instruction_table = channel_description.instruction_table
    for instruction in instruction_table:
        instruction_type = type(instruction.operation)
        if instruction_type not in device_constraints.compatible_instructions:
            raise AWGScheduleValidationError(
                f"Incompatible instruction {instruction_type.__name__} not compatible with "
                f"{channel_description.controller_name}"
            )
        if instruction_type == ConditionalInstruction:
            if_false_duration = instruction.operation.if_false.duration_samples
            if_true_duration = instruction.operation.if_true.duration_samples
            if if_false_duration != instruction.duration_samples or if_true_duration != instruction.duration_samples:
                raise AWGScheduleValidationError(
                    "Conditional Instruction's instructions must have the same duration\n"
                    f"Duration of Conditional instruction: {instruction.duration_samples} \n"
                    f"Duration of if_false instruction: {if_false_duration}\n"
                    f"Duration of if_true instruction: {if_true_duration}\n"
                )
        elif instruction_type == ReadoutTrigger:
            if instruction.duration_samples <= instruction.operation.probe_pulse.duration_samples:
                raise AWGScheduleValidationError(
                    "Duration of ReadoutTrigger must be longer than the probe pulse inside it."
                )

        if (instruction.duration_samples % device_constraints.granularity) != 0:
            raise AWGScheduleValidationError(
                f"Instruction length of {instruction.duration_samples} doesn't match the "
                f"{channel_description.controller_name} granularity of the device"
            )
        if instruction.duration_samples < device_constraints.min_number_of_samples:
            raise AWGScheduleValidationError(
                f"Instruction length of {instruction.duration_samples} is less than the minimum number "
                f"of samples {device_constraints.min_number_of_samples} for {channel_description.controller_name}"
            )
        validate_instruction_and_wf_length(instruction)
